@@ -42,6 +42,8 @@
 #include "gok-word-complete.h"
 #include "gok-keyboard.h"
 
+#define GOK_MAX_SEARCH_DEPTH 18
+
 /*
  * NOTE: we can tweak the efficiency/accuracy of the
  * gui search by adjusting using the spy/gui_search_depth and 
@@ -134,7 +136,7 @@ static gboolean spyshutdown = FALSE;/* this global could be removed if we have a
 						way of removing our idle handler in gok_spy_stop */
 
 #define MAX_CHILD_COUNT 100
-
+#define MAX_NAME_SEARCH_DEPTH 10
 
 static void
 gok_spy_set_focussed_object (Accessible *accessible)
@@ -720,34 +722,53 @@ gok_spy_get_list_parent (Accessible *search_root)
 }
 
 static gchar*
-gok_spy_concatenate_child_names (Accessible *parent)
+gok_spy_concatenate_child_names (Accessible *parent, gint depth)
 {
 	gint num_children = 0;
 	gchar *tmp = NULL, *name = NULL;
+
+	/* avoid cyclic graph */
+	if (depth++ >= MAX_NAME_SEARCH_DEPTH) 
+	{
+		return NULL;
+	}
+		
 	if (parent) 
 	{
 		gint i;
 		num_children = Accessible_getChildCount (parent);
-		for (i = 0; i < num_children; ++i) 
+		for (i = 0; i < num_children; ++i)  
 		{
 			Accessible *child = Accessible_getChildAtIndex (parent, i);
 			if (child ) 
 			{
-				gchar *childname = Accessible_getName (child);
+				gchar *childname;
+				gchar *spi_childname = Accessible_getName (child);
+
+				if (spi_childname && (strlen (spi_childname) > 0))  
+				{
+					childname = g_strdup (spi_childname);
+				}
+				else 
+				{
+					childname = gok_spy_concatenate_child_names (child, depth);
+				}
 				if (childname && (strlen (childname) > 0)) 
 				{
-					if (tmp)
+					if (tmp) 
 					{
 						name = g_strconcat (tmp, "", childname, NULL);
 						g_free (tmp);
 					}
-					else
+					else 
 					{
 						name = g_strdup (childname);
 					}
-					SPI_freeString (childname);
 					tmp = name;
 				}
+
+				SPI_freeString (spi_childname);	
+				g_free (childname);
 			}
 		}
 	}
@@ -1634,7 +1655,7 @@ GSList* gok_spy_append_node (GSList* nodes,
 		 * is the dominant user activity, this is especially appropriate.
 		 */
 		if (pName == NULL && role == SPI_ROLE_TABLE_CELL) {
-			pName = gok_spy_concatenate_child_names (pAccessible);
+			pName = gok_spy_concatenate_child_names (pAccessible, 0);
 		}
 		else if (pName == NULL && role == SPI_ROLE_SCROLL_BAR) {
 		    states = Accessible_getStateSet (pAccessible);
@@ -2026,6 +2047,7 @@ void gok_spy_process_focus (Accessible* accessible)
 	Accessible* pnewWindowAccessible;
 	Accessible* ptempAccessible;
 	gboolean has_menu = FALSE;
+	AccessibleRole role;
 	
 	gok_log_enter();
 	g_assert(accessible != NULL);
@@ -2084,12 +2106,18 @@ void gok_spy_process_focus (Accessible* accessible)
 	 * if current active window is NULL, and this is a menu item, 
 	 * then we're in a popup menu of some kind; branch to the menus kbd
 	 */
+	role = Accessible_getRole (accessible);
 
 	if (accessible && (m_ptheWindowAccessible == NULL) && 
-	    gok_spy_is_menu_role (Accessible_getRole (accessible)))
+	    gok_spy_is_menu_role (role))
 	{
 	    AccessibleNode *node = g_new0 (AccessibleNode, 1);
-	    gchar *name = Accessible_getName (accessible);
+	    gchar *name;
+	    if (role == SPI_ROLE_MENU_ITEM)
+	    {
+		accessible = Accessible_getParent (accessible);
+	    }
+	    name = Accessible_getName (accessible);
 	    node->link = -1;
 	    node->flags.value = 0;
 	    node->flags.data.is_menu = TRUE;
@@ -2308,7 +2336,7 @@ gok_spy_find_and_append_toolbar_items (Accessible *root, AccessibleNodeFlags fla
 }
 
 static GokSpyUIFlags 
-gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboard_ui_flags, AccessibleNodeFlags context_flags)
+gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboard_ui_flags, AccessibleNodeFlags context_flags, gint search_depth)
 {
 	Accessible* child;
 	Accessible* parent;
@@ -2323,6 +2351,13 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 	ui_flags.value = 0;
 	parent = rootAccessible;
 
+	if (++search_depth > GOK_MAX_SEARCH_DEPTH) 
+	{
+		g_warning ("GOK: maximum search depth exceeded.");
+		gok_log_leave ();
+		return ui_flags;
+	}			
+
 	if (gok_spy_worth_searching (parent) == FALSE) 
 	{
 		gok_log ("not worth searching.");
@@ -2347,6 +2382,9 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 		flags.data.inside_html_container = context_flags.data.inside_html_container;
 
 		child = Accessible_getChildAtIndex (parent, i);
+		if (child == parent) /* bogus hierarchy, but yelp sometimes does this */
+		    continue;
+
 		if (child)
 		{
 			gok_spy_accessible_implicit_ref (child);
@@ -2369,7 +2407,7 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				ui_flags.data.toolbars = flags.data.is_toolbar_item = TRUE;
 				if (!gok_spy_find_and_append_toolbar_items (child, flags)) 
 				{
-				    ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
+				    ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
 				}
 			}
 			if ((keyboard_ui_flags.data.gui || keyboard_ui_flags.data.editable_text) && 
@@ -2395,12 +2433,12 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				    /* special case for notebook tabs and expanders, whose children should also appear */
   				    if (role == SPI_ROLE_PAGE_TAB || role == SPI_ROLE_TOGGLE_BUTTON)
   				    {
-  					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
+  					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
   				    }
 				}
 				else
 				{
-					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
+					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
 				}
 			}
 			gok_spy_accessible_unref (child);
@@ -2442,7 +2480,8 @@ gok_spy_update_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 	_priv_ui_nodes = NULL;
 
 	context_flags.value = 0;
-	ui_flags = gok_spy_search_component_list (rootAccessible, keyboard_ui_flags, context_flags);
+	ui_flags = gok_spy_search_component_list (rootAccessible, keyboard_ui_flags, context_flags, 0);
+
 	/* _priv_ui_nodes is now re-assigned, by above call. */
 	gok_spy_resolve_namesakes (_priv_ui_nodes);
 
